@@ -1,9 +1,10 @@
 /*
- * Copyright  2000-2004 The Apache Software Foundation
- *
- *  Licensed under the Apache License, Version 2.0 (the "License");
- *  you may not use this file except in compliance with the License.
- *  You may obtain a copy of the License at
+ *  Licensed to the Apache Software Foundation (ASF) under one or more
+ *  contributor license agreements.  See the NOTICE file distributed with
+ *  this work for additional information regarding copyright ownership.
+ *  The ASF licenses this file to You under the Apache License, Version 2.0
+ *  (the "License"); you may not use this file except in compliance with
+ *  the License.  You may obtain a copy of the License at
  *
  *      http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -14,37 +15,24 @@
  *  limitations under the License.
  *
  */
-
 package org.apache.tools.ant;
 
-import java.io.BufferedReader;
 import java.io.File;
-import java.io.InputStream;
-import java.io.InputStreamReader;
 import java.util.Hashtable;
 import java.util.Locale;
 import java.util.Vector;
-import org.apache.tools.ant.helper.ProjectHelper2;
+
+import org.apache.tools.ant.types.Resource;
+import org.apache.tools.ant.types.resources.FileResource;
 import org.apache.tools.ant.util.LoaderUtils;
 import org.xml.sax.AttributeList;
 
 /**
  * Configures a Project (complete with Targets and Tasks) based on
- * a XML build file. It'll rely on a plugin to do the actual processing
- * of the xml file.
- *
+ * a build file. It'll rely on a plugin to do the actual processing
+ * of the file.
+ * <p>
  * This class also provide static wrappers for common introspection.
- *
- * All helper plugins must provide backward compatibility with the
- * original ant patterns, unless a different behavior is explicitly
- * specified. For example, if namespace is used on the &lt;project&gt; tag
- * the helper can expect the entire build file to be namespace-enabled.
- * Namespaces or helper-specific tags can provide meta-information to
- * the helper, allowing it to use new ( or different policies ).
- *
- * However, if no namespace is used the behavior should be exactly
- * identical with the default helper.
- *
  */
 public class ProjectHelper {
     /** The URI for ant name space */
@@ -63,31 +51,32 @@ public class ProjectHelper {
      * Name of JVM system property which provides the name of the
      * ProjectHelper class to use.
      */
-    public static final String HELPER_PROPERTY =
-        "org.apache.tools.ant.ProjectHelper";
+    public static final String HELPER_PROPERTY = MagicNames.PROJECT_HELPER_CLASS;
 
     /**
      * The service identifier in jars which provide Project Helper
      * implementations.
      */
-    public static final String SERVICE_ID =
-        "META-INF/services/org.apache.tools.ant.ProjectHelper";
+    public static final String SERVICE_ID = MagicNames.PROJECT_HELPER_SERVICE;
 
     /**
-     * Configures the project with the contents of the specified XML file.
+     * name of project helper reference that we add to a project
+     */
+    public static final String PROJECTHELPER_REFERENCE = MagicNames.REFID_PROJECT_HELPER;
+
+    /**
+     * Configures the project with the contents of the specified build file.
      *
      * @param project The project to configure. Must not be <code>null</code>.
-     * @param buildFile An XML file giving the project's configuration.
+     * @param buildFile A build file giving the project's configuration.
      *                  Must not be <code>null</code>.
      *
-     * @deprecated Use the non-static parse method
-     * @exception BuildException if the configuration is invalid or cannot
-     *                           be read
+     * @exception BuildException if the configuration is invalid or cannot be read
      */
-    public static void configureProject(Project project, File buildFile)
-        throws BuildException {
-        ProjectHelper helper = ProjectHelper.getProjectHelper();
-        project.addReference("ant.projectHelper", helper);
+    public static void configureProject(Project project, File buildFile) throws BuildException {
+        FileResource resource = new FileResource(buildFile);
+        ProjectHelper helper = ProjectHelperRepository.getInstance().getProjectHelperForBuildFile(resource);
+        project.addReference(PROJECTHELPER_REFERENCE, helper);
         helper.parse(project, buildFile);
     }
 
@@ -99,24 +88,9 @@ public class ProjectHelper {
     // The following properties are required by import ( and other tasks
     // that read build files using ProjectHelper ).
 
-    // A project helper may process multiple files. We'll keep track
-    // of them - to avoid loops and to allow caching. The caching will
-    // probably accelerate things like <antCall>.
-    // The key is the absolute file, the value is a processed tree.
-    // Since the tree is composed of UE and RC - it can be reused !
-    // protected Hashtable processedFiles=new Hashtable();
-
     private Vector importStack = new Vector();
 
-    // Temporary - until we figure a better API
-    /** EXPERIMENTAL WILL_CHANGE
-     *
-     */
-//    public Hashtable getProcessedFiles() {
-//        return processedFiles;
-//    }
-
-    /** EXPERIMENTAL WILL_CHANGE
+    /**
      *  Import stack.
      *  Used to keep track of imported files. Error reporting should
      *  display the import path.
@@ -127,6 +101,94 @@ public class ProjectHelper {
         return importStack;
     }
 
+    private final static ThreadLocal targetPrefix = new ThreadLocal() {
+            protected Object initialValue() {
+                return (String) null;
+            }
+        };
+
+    /**
+     * The prefix to prepend to imported target names.
+     *
+     * <p>May be set by &lt;import&gt;'s as attribute.</p>
+     *
+     * @return the configured prefix or null
+     *
+     * @since Ant 1.8.0
+     */
+    public static String getCurrentTargetPrefix() {
+        return (String) targetPrefix.get();
+    }
+
+    /**
+     * Sets the prefix to prepend to imported target names.
+     *
+     * @since Ant 1.8.0
+     */
+    public static void setCurrentTargetPrefix(String prefix) {
+        targetPrefix.set(prefix);
+    }
+
+    private final static ThreadLocal prefixSeparator = new ThreadLocal() {
+            protected Object initialValue() {
+                return ".";
+            }
+        };
+
+    /**
+     * The separator between the prefix and the target name.
+     *
+     * <p>May be set by &lt;import&gt;'s prefixSeperator attribute.</p>
+     *
+     * @since Ant 1.8.0
+     */
+    public static String getCurrentPrefixSeparator() {
+        return (String) prefixSeparator.get();
+    }
+
+    /**
+     * Sets the separator between the prefix and the target name.
+     *
+     * @since Ant 1.8.0
+     */
+    public static void setCurrentPrefixSeparator(String sep) {
+        prefixSeparator.set(sep);
+    }
+
+    private final static ThreadLocal inIncludeMode = new ThreadLocal() {
+            protected Object initialValue() {
+                return Boolean.FALSE;
+            }
+        };
+
+    /**
+     * Whether the current file should be read in include as opposed
+     * to import mode.
+     *
+     * <p>In include mode included targets are only known by their
+     * prefixed names and their depends lists get rewritten so that
+     * all dependencies get the prefix as well.</p>
+     *
+     * <p>In import mode imported targets are known by an adorned as
+     * well as a prefixed name and the unadorned target may be
+     * overwritten in the importing build file.  The depends list of
+     * the imported targets is not modified at all.</p>
+     *
+     * @since Ant 1.8.0
+     */
+    public static boolean isInIncludeMode() {
+        return inIncludeMode.get() == Boolean.TRUE;
+    }
+
+    /**
+     * Sets whether the current file should be read in include as
+     * opposed to import mode.
+     *
+     * @since Ant 1.8.0
+     */
+    public static void setInIncludeMode(boolean includeMode) {
+        inIncludeMode.set(includeMode ? Boolean.TRUE : Boolean.FALSE);
+    }
 
     // --------------------  Parse method  --------------------
     /**
@@ -147,139 +209,27 @@ public class ProjectHelper {
             + "in a helper plugin " + this.getClass().getName());
     }
 
-
     /**
-     * Discovers a project helper instance. Uses the same patterns
-     * as JAXP, commons-logging, etc: a system property, a JDK1.3
-     * service discovery, default.
-     *
-     * @return a ProjectHelper, either a custom implementation
-     * if one is available and configured, or the default implementation
-     * otherwise.
-     *
-     * @exception BuildException if a specified helper class cannot
-     * be loaded/instantiated.
+     * Get the first project helper found in the classpath
+     * 
+     * @return an project helper, never <code>null</code>
+     * @see #getHelpers()
      */
-    public static ProjectHelper getProjectHelper()
-        throws BuildException {
-        // Identify the class loader we will be using. Ant may be
-        // in a webapp or embedded in a different app
-        ProjectHelper helper = null;
-
-        // First, try the system property
-        String helperClass = System.getProperty(HELPER_PROPERTY);
-        try {
-            if (helperClass != null) {
-                helper = newHelper(helperClass);
-            }
-        } catch (SecurityException e) {
-            System.out.println("Unable to load ProjectHelper class \""
-                + helperClass + " specified in system property "
-                + HELPER_PROPERTY);
-        }
-
-        // A JDK1.3 'service' ( like in JAXP ). That will plug a helper
-        // automatically if in CLASSPATH, with the right META-INF/services.
-        if (helper == null) {
-            try {
-                ClassLoader classLoader = LoaderUtils.getContextClassLoader();
-                InputStream is = null;
-                if (classLoader != null) {
-                    is = classLoader.getResourceAsStream(SERVICE_ID);
-                }
-                if (is == null) {
-                    is = ClassLoader.getSystemResourceAsStream(SERVICE_ID);
-                }
-
-                if (is != null) {
-                    // This code is needed by EBCDIC and other strange systems.
-                    // It's a fix for bugs reported in xerces
-                    InputStreamReader isr;
-                    try {
-                        isr = new InputStreamReader(is, "UTF-8");
-                    } catch (java.io.UnsupportedEncodingException e) {
-                        isr = new InputStreamReader(is);
-                    }
-                    BufferedReader rd = new BufferedReader(isr);
-
-                    String helperClassName = rd.readLine();
-                    rd.close();
-
-                    if (helperClassName != null
-                        && !"".equals(helperClassName)) {
-
-                        helper = newHelper(helperClassName);
-                    }
-                }
-            } catch (Exception ex) {
-                System.out.println("Unable to load ProjectHelper "
-                    + "from service \"" + SERVICE_ID);
-            }
-        }
-
-        if (helper != null) {
-            return helper;
-        } else {
-            try {
-                // Default
-                // return new ProjectHelperImpl();
-                return new ProjectHelper2();
-            } catch (Throwable e) {
-                String message = "Unable to load default ProjectHelper due to "
-                    + e.getClass().getName() + ": " + e.getMessage();
-                throw new BuildException(message, e);
-            }
-        }
+    public static ProjectHelper getProjectHelper() {
+        return (ProjectHelper) ProjectHelperRepository.getInstance().getHelpers().next();
     }
 
     /**
-     * Creates a new helper instance from the name of the class.
-     * It'll first try the thread class loader, then Class.forName()
-     * will load from the same loader that loaded this class.
+     * JDK1.1 compatible access to the context class loader. Cut & paste from JAXP.
      *
-     * @param helperClass The name of the class to create an instance
-     *                    of. Must not be <code>null</code>.
+     * @deprecated since 1.6.x.
+     *             Use LoaderUtils.getContextClassLoader()
      *
-     * @return a new instance of the specified class.
-     *
-     * @exception BuildException if the class cannot be found or
-     * cannot be appropriate instantiated.
-     */
-    private static ProjectHelper newHelper(String helperClass)
-        throws BuildException {
-        ClassLoader classLoader = LoaderUtils.getContextClassLoader();
-        try {
-            Class clazz = null;
-            if (classLoader != null) {
-                try {
-                    clazz = classLoader.loadClass(helperClass);
-                } catch (ClassNotFoundException ex) {
-                    // try next method
-                }
-            }
-            if (clazz == null) {
-                clazz = Class.forName(helperClass);
-            }
-            return ((ProjectHelper) clazz.newInstance());
-        } catch (Exception e) {
-            throw new BuildException(e);
-        }
-    }
-
-    /**
-     * JDK1.1 compatible access to the context class loader.
-     * Cut&paste from JAXP.
-     *
-     * @deprecated Use LoaderUtils.getContextClassLoader()
      * @return the current context class loader, or <code>null</code>
      * if the context class loader is unavailable.
      */
     public static ClassLoader getContextClassLoader() {
-        if (!LoaderUtils.isContextLoaderAvailable()) {
-            return null;
-        }
-
-        return LoaderUtils.getContextClassLoader();
+        return LoaderUtils.isContextLoaderAvailable() ? LoaderUtils.getContextClassLoader() : null;
     }
 
     // -------------------- Static utils, used by most helpers ----------------
@@ -294,7 +244,9 @@ public class ProjectHelper {
      * @param project The project containing the target.
      *                Must not be <code>null</code>.
      *
-     * @deprecated Use IntrospectionHelper for each property
+     * @deprecated since 1.6.x.
+     *             Use IntrospectionHelper for each property.
+     *
      * @exception BuildException if any of the attributes can't be handled by
      *                           the target
      */
@@ -303,20 +255,13 @@ public class ProjectHelper {
         if (target instanceof TypeAdapter) {
             target = ((TypeAdapter) target).getProxy();
         }
+        IntrospectionHelper ih = IntrospectionHelper.getHelper(project, target.getClass());
 
-        IntrospectionHelper ih =
-            IntrospectionHelper.getHelper(target.getClass());
-
-        project.addBuildListener(ih);
-
-        for (int i = 0; i < attrs.getLength(); i++) {
+        for (int i = 0, length = attrs.getLength(); i < length; i++) {
             // reflect these into the target
-            String value = replaceProperties(project, attrs.getValue(i),
-                                             project.getProperties());
+            String value = replaceProperties(project, attrs.getValue(i), project.getProperties());
             try {
-                ih.setAttribute(project, target,
-                                attrs.getName(i).toLowerCase(Locale.US), value);
-
+                ih.setAttribute(project, target, attrs.getName(i).toLowerCase(Locale.US), value);
             } catch (BuildException be) {
                 // id attribute must be set externally
                 if (!attrs.getName(i).equals("id")) {
@@ -364,13 +309,10 @@ public class ProjectHelper {
         if (text == null) {
             return;
         }
-
         if (target instanceof TypeAdapter) {
             target = ((TypeAdapter) target).getProxy();
         }
-
-        IntrospectionHelper.getHelper(target.getClass()).addText(project,
-            target, text);
+        IntrospectionHelper.getHelper(project, target.getClass()).addText(project, target, text);
     }
 
     /**
@@ -386,10 +328,8 @@ public class ProjectHelper {
      *                May be <code>null</code>, in which case
      *                the child is not stored.
      */
-    public static void storeChild(Project project, Object parent,
-         Object child, String tag) {
-        IntrospectionHelper ih
-            = IntrospectionHelper.getHelper(parent.getClass());
+    public static void storeChild(Project project, Object parent, Object child, String tag) {
+        IntrospectionHelper ih = IntrospectionHelper.getHelper(project, parent.getClass());
         ih.storeElement(project, parent, child, tag);
     }
 
@@ -409,11 +349,11 @@ public class ProjectHelper {
      * @return the original string with the properties replaced, or
      *         <code>null</code> if the original string is <code>null</code>.
      *
-     * @deprecated Use project.replaceProperties()
+     * @deprecated since 1.6.x.
+     *             Use project.replaceProperties().
      * @since 1.5
      */
-     public static String replaceProperties(Project project, String value)
-            throws BuildException {
+     public static String replaceProperties(Project project, String value) throws BuildException {
         // needed since project properties are not accessible
          return project.replaceProperties(value);
      }
@@ -435,10 +375,11 @@ public class ProjectHelper {
      *                           <code>}</code>
      * @return the original string with the properties replaced, or
      *         <code>null</code> if the original string is <code>null</code>.
-     * @deprecated Use PropertyHelper
+     * @deprecated since 1.6.x.
+     *             Use PropertyHelper.
      */
-     public static String replaceProperties(Project project, String value,
-         Hashtable keys) throws BuildException {
+     public static String replaceProperties(Project project, String value, Hashtable keys)
+             throws BuildException {
         PropertyHelper ph = PropertyHelper.getPropertyHelper(project);
         return ph.replaceProperties(null, value, keys);
     }
@@ -450,23 +391,25 @@ public class ProjectHelper {
      * <code>null</code> entries in the first list indicate a property
      * reference from the second list.
      *
+     * <p>As of Ant 1.8.0 this method is never invoked by any code
+     * inside of Ant itself.</p>
+     *
      * @param value     Text to parse. Must not be <code>null</code>.
      * @param fragments List to add text fragments to.
      *                  Must not be <code>null</code>.
      * @param propertyRefs List to add property names to.
      *                     Must not be <code>null</code>.
      *
-     * @deprecated Use PropertyHelper
+     * @deprecated since 1.6.x.
+     *             Use PropertyHelper.
      * @exception BuildException if the string contains an opening
-     *                           <code>${</code> without a closing
-     *                           <code>}</code>
+     *                           <code>${</code> without a closing <code>}</code>
      */
-    public static void parsePropertyString(String value, Vector fragments,
-                                           Vector propertyRefs)
-        throws BuildException {
-        PropertyHelper.parsePropertyStringDefault(value, fragments,
-                propertyRefs);
+    public static void parsePropertyString(String value, Vector fragments, Vector propertyRefs)
+            throws BuildException {
+        PropertyHelper.parsePropertyStringDefault(value, fragments, propertyRefs);
     }
+
     /**
      * Map a namespaced {uri,name} to an internal string format.
      * For BC purposes the names from the ant core uri will be
@@ -524,7 +467,7 @@ public class ProjectHelper {
      *         did not have a location, just return the build exception
      */
     public static BuildException addLocationToBuildException(
-        BuildException ex, Location newLocation) {
+            BuildException ex, Location newLocation) {
         if (ex.getLocation() == null || ex.getMessage() == null) {
             return ex;
         }
@@ -535,9 +478,63 @@ public class ProjectHelper {
             + ex.getMessage();
         if (newLocation == null) {
             return new BuildException(errorMessage, ex);
-        } else {
-            return new BuildException(
-                errorMessage, ex, newLocation);
         }
+        return new BuildException(errorMessage, ex, newLocation);
+    }
+
+    /**
+     * Whether this instance of ProjectHelper can parse an Antlib
+     * descriptor given by the URL and return its content as an
+     * UnknownElement ready to be turned into an Antlib task.
+     *
+     * <p>This method should not try to parse the content of the
+     * descriptor, the URL is only given as an argument to allow
+     * subclasses to decide whether they can support a given URL
+     * scheme or not.</p>
+     *
+     * <p>Subclasses that return true in this method must also
+     * override {@link #parseAntlibDescriptor
+     * parseAntlibDescriptor}.</p>
+     *
+     * <p>This implementation returns false.</p>
+     *
+     * @since Ant 1.8.0
+     */
+    public boolean canParseAntlibDescriptor(Resource r) {
+        return false;
+    }
+
+    /**
+     * Parse the given URL as an antlib descriptor and return the
+     * content as something that can be turned into an Antlib task.
+     *
+     * @since ant 1.8.0
+     */
+    public UnknownElement parseAntlibDescriptor(Project containingProject,
+                                                Resource source) {
+        throw new BuildException("can't parse antlib descriptors");
+    }
+
+    /**
+     * Check if the helper supports the kind of file. Some basic check on the
+     * extension's file should be done here.
+     * 
+     * @param buildFile
+     *            the file expected to be parsed (never <code>null</code>)
+     * @return true if the helper supports it
+     * @since Ant 1.8.0
+     */
+    public boolean canParseBuildFile(Resource buildFile) {
+        return true;
+    }
+
+    /**
+     * The file name of the build script to be parsed if none specified on the command line
+     * 
+     * @return the name of the default file (never <code>null</code>)
+     * @since Ant 1.8.0
+     */
+    public String getDefaultBuildFile() {
+        return Main.DEFAULT_BUILD_FILENAME;
     }
 }
