@@ -1,9 +1,10 @@
 /*
- * Copyright  2000-2004 The Apache Software Foundation
- *
- *  Licensed under the Apache License, Version 2.0 (the "License");
- *  you may not use this file except in compliance with the License.
- *  You may obtain a copy of the License at
+ *  Licensed to the Apache Software Foundation (ASF) under one or more
+ *  contributor license agreements.  See the NOTICE file distributed with
+ *  this work for additional information regarding copyright ownership.
+ *  The ASF licenses this file to You under the Apache License, Version 2.0
+ *  (the "License"); you may not use this file except in compliance with
+ *  the License.  You may obtain a copy of the License at
  *
  *      http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -18,30 +19,48 @@
 package org.apache.tools.ant.taskdefs;
 
 import java.io.File;
+import java.util.Arrays;
 import java.util.Vector;
+import java.util.Iterator;
+import java.util.Comparator;
+
+import org.apache.tools.ant.Project;
 import org.apache.tools.ant.BuildException;
 import org.apache.tools.ant.DirectoryScanner;
-import org.apache.tools.ant.Project;
 import org.apache.tools.ant.taskdefs.condition.Os;
 import org.apache.tools.ant.types.FileSet;
 import org.apache.tools.ant.types.PatternSet;
+import org.apache.tools.ant.types.Resource;
+import org.apache.tools.ant.types.ResourceCollection;
+import org.apache.tools.ant.types.resources.FileProvider;
+import org.apache.tools.ant.types.resources.Sort;
+import org.apache.tools.ant.types.resources.Restrict;
+import org.apache.tools.ant.types.resources.Resources;
+import org.apache.tools.ant.types.resources.FileResourceIterator;
+import org.apache.tools.ant.types.resources.comparators.Reverse;
+import org.apache.tools.ant.types.resources.comparators.FileSystem;
+import org.apache.tools.ant.types.resources.comparators.ResourceComparator;
+import org.apache.tools.ant.types.resources.selectors.Exists;
+import org.apache.tools.ant.types.resources.selectors.ResourceSelector;
+import org.apache.tools.ant.types.selectors.OrSelector;
 import org.apache.tools.ant.types.selectors.AndSelector;
-import org.apache.tools.ant.types.selectors.ContainsRegexpSelector;
-import org.apache.tools.ant.types.selectors.ContainsSelector;
+import org.apache.tools.ant.types.selectors.NotSelector;
 import org.apache.tools.ant.types.selectors.DateSelector;
-import org.apache.tools.ant.types.selectors.DependSelector;
+import org.apache.tools.ant.types.selectors.FileSelector;
+import org.apache.tools.ant.types.selectors.NoneSelector;
+import org.apache.tools.ant.types.selectors.SizeSelector;
 import org.apache.tools.ant.types.selectors.DepthSelector;
+import org.apache.tools.ant.types.selectors.DependSelector;
 import org.apache.tools.ant.types.selectors.ExtendSelector;
+import org.apache.tools.ant.types.selectors.SelectSelector;
+import org.apache.tools.ant.types.selectors.PresentSelector;
+import org.apache.tools.ant.types.selectors.ContainsSelector;
 import org.apache.tools.ant.types.selectors.FilenameSelector;
 import org.apache.tools.ant.types.selectors.MajoritySelector;
-import org.apache.tools.ant.types.selectors.NoneSelector;
-import org.apache.tools.ant.types.selectors.NotSelector;
-import org.apache.tools.ant.types.selectors.OrSelector;
-import org.apache.tools.ant.types.selectors.PresentSelector;
-import org.apache.tools.ant.types.selectors.SelectSelector;
-import org.apache.tools.ant.types.selectors.SizeSelector;
-import org.apache.tools.ant.types.selectors.FileSelector;
+import org.apache.tools.ant.types.selectors.ContainsRegexpSelector;
 import org.apache.tools.ant.types.selectors.modifiedselector.ModifiedSelector;
+import org.apache.tools.ant.util.FileUtils;
+import org.apache.tools.ant.util.SymbolicLinkUtils;
 
 /**
  * Deletes a file or directory, or set of files defined by a fileset.
@@ -49,7 +68,7 @@ import org.apache.tools.ant.types.selectors.modifiedselector.ModifiedSelector;
  * using the include/exclude syntax.  The deltree task would delete a
  * directory tree.  This task combines the functionality of these two
  * originally distinct tasks.
- * <p>Currently Delete extends MatchingTask.  This is intend <i>only</i>
+ * <p>Currently Delete extends MatchingTask.  This is intended <i>only</i>
  * to provide backwards compatibility for a release.  The future position
  * is to use nested filesets exclusively.</p>
  *
@@ -58,18 +77,49 @@ import org.apache.tools.ant.types.selectors.modifiedselector.ModifiedSelector;
  * @ant.task category="filesystem"
  */
 public class Delete extends MatchingTask {
-    private static final int DELETE_RETRY_SLEEP_MILLIS = 10;
+    private static final ResourceComparator REVERSE_FILESYSTEM = new Reverse(new FileSystem());
+    private static final ResourceSelector EXISTS = new Exists();
+
+    private static class ReverseDirs implements ResourceCollection {
+        static final Comparator REVERSE = new Comparator() {
+            public int compare(Object foo, Object bar) {
+                return ((Comparable) foo).compareTo(bar) * -1;
+            }
+        };
+        private Project project;
+        private File basedir;
+        private String[] dirs;
+        ReverseDirs(Project project, File basedir, String[] dirs) {
+            this.project = project;
+            this.basedir = basedir;
+            this.dirs = dirs;
+            Arrays.sort(this.dirs, REVERSE);
+        }
+        public Iterator iterator() {
+            return new FileResourceIterator(project, basedir, dirs);
+        }
+        public boolean isFilesystemOnly() { return true; }
+        public int size() { return dirs.length; }
+    }
+
+    // CheckStyle:VisibilityModifier OFF - bc
     protected File file = null;
     protected File dir = null;
     protected Vector filesets = new Vector();
     protected boolean usedMatchingTask = false;
     // by default, remove matching empty dirs
     protected boolean includeEmpty = false;
+    // CheckStyle:VisibilityModifier ON
 
     private int verbosity = Project.MSG_VERBOSE;
     private boolean quiet = false;
     private boolean failonerror = true;
     private boolean deleteOnExit = false;
+    private boolean removeNotFollowedSymlinks = false;
+    private Resources rcs = null;
+    private static FileUtils FILE_UTILS = FileUtils.getFileUtils();
+    private static SymbolicLinkUtils SYMLINK_UTILS =
+        SymbolicLinkUtils.getSymbolicLinkUtils();
 
     /**
      * Set the name of a single file to be removed.
@@ -87,6 +137,7 @@ public class Delete extends MatchingTask {
      */
     public void setDir(File dir) {
         this.dir = dir;
+        getImplicitFileSet().setDir(dir);
     }
 
     /**
@@ -155,6 +206,21 @@ public class Delete extends MatchingTask {
         filesets.addElement(set);
     }
 
+    /**
+     * Add an arbitrary ResourceCollection to be deleted.
+     * @param rc the filesystem-only ResourceCollection.
+     */
+    public void add(ResourceCollection rc) {
+        if (rc == null) {
+            return;
+        }
+        if (rcs == null) {
+            rcs = new Resources();
+            rcs.setCache(true);
+        }
+        rcs.add(rc);
+    }
+
     /**
      * add a name entry on the include list
      * @return a NameEntry object to be configured
@@ -277,6 +343,16 @@ public class Delete extends MatchingTask {
         super.setFollowSymlinks(followSymlinks);
     }
 
+    /**
+     * Sets whether the symbolic links that have not been followed
+     * shall be removed (the links, not the locations they point at).
+     *
+     * @since Ant 1.8.0
+     */
+    public void setRemoveNotFollowedSymlinks(boolean b) {
+        removeNotFollowedSymlinks = b;
+    }
+
     /**
      * add a "Select" selector entry on the selector list
      * @param selector the selector to be added
@@ -439,12 +515,12 @@ public class Delete extends MatchingTask {
     public void execute() throws BuildException {
         if (usedMatchingTask) {
             log("DEPRECATED - Use of the implicit FileSet is deprecated.  "
-                + "Use a nested fileset element instead.");
+                + "Use a nested fileset element instead.", quiet ? Project.MSG_VERBOSE : verbosity);
         }
 
-        if (file == null && dir == null && filesets.size() == 0) {
+        if (file == null && dir == null && filesets.size() == 0 && rcs == null) {
             throw new BuildException("At least one of the file or dir "
-                                     + "attributes, or a fileset element, "
+                                     + "attributes, or a nested resource collection, "
                                      + "must be set.");
         }
 
@@ -453,85 +529,166 @@ public class Delete extends MatchingTask {
                                      + "set to true", getLocation());
         }
 
-
         // delete the single file
         if (file != null) {
             if (file.exists()) {
                 if (file.isDirectory()) {
                     log("Directory " + file.getAbsolutePath()
                         + " cannot be removed using the file attribute.  "
-                        + "Use dir instead.");
+                        + "Use dir instead.", quiet ? Project.MSG_VERBOSE : verbosity);
                 } else {
                     log("Deleting: " + file.getAbsolutePath());
 
                     if (!delete(file)) {
-                        String message = "Unable to delete file "
-                            + file.getAbsolutePath();
-                        if (failonerror) {
-                            throw new BuildException(message);
-                        } else {
-                            log(message, quiet ? Project.MSG_VERBOSE
-                                               : Project.MSG_WARN);
-                        }
+                        handle("Unable to delete file " + file.getAbsolutePath());
                     }
                 }
+            } else if (isDanglingSymlink(file)) {
+                log("Trying to delete file " + file.getAbsolutePath()
+                    + " which looks like a broken symlink.",
+                    quiet ? Project.MSG_VERBOSE : verbosity);
+                if (!delete(file)) {
+                    handle("Unable to delete file " + file.getAbsolutePath());
+                }
             } else {
                 log("Could not find file " + file.getAbsolutePath()
-                    + " to delete.",
-                    Project.MSG_VERBOSE);
+                    + " to delete.", quiet ? Project.MSG_VERBOSE : verbosity);
             }
         }
 
         // delete the directory
-        if (dir != null && dir.exists() && dir.isDirectory()
-            && !usedMatchingTask) {
-            /*
-               If verbosity is MSG_VERBOSE, that mean we are doing
-               regular logging (backwards as that sounds).  In that
-               case, we want to print one message about deleting the
-               top of the directory tree.  Otherwise, the removeDir
-               method will handle messages for _all_ directories.
-             */
-            if (verbosity == Project.MSG_VERBOSE) {
-                log("Deleting directory " + dir.getAbsolutePath());
-            }
-            removeDir(dir);
-        }
-
-        // delete the files in the filesets
-        for (int i = 0; i < filesets.size(); i++) {
-            FileSet fs = (FileSet) filesets.elementAt(i);
-            try {
-                DirectoryScanner ds = fs.getDirectoryScanner(getProject());
-                String[] files = ds.getIncludedFiles();
-                String[] dirs = ds.getIncludedDirectories();
-                removeFiles(fs.getDir(getProject()), files, dirs);
-            } catch (BuildException be) {
-                // directory doesn't exist or is not readable
-                if (failonerror) {
-                    throw be;
-                } else {
-                    log(be.getMessage(),
-                        quiet ? Project.MSG_VERBOSE : Project.MSG_WARN);
+        if (dir != null && !usedMatchingTask) {
+            if (dir.exists() && dir.isDirectory()) {
+                /*
+                  If verbosity is MSG_VERBOSE, that mean we are doing
+                  regular logging (backwards as that sounds).  In that
+                  case, we want to print one message about deleting the
+                  top of the directory tree.  Otherwise, the removeDir
+                  method will handle messages for _all_ directories.
+                */
+                if (verbosity == Project.MSG_VERBOSE) {
+                    log("Deleting directory " + dir.getAbsolutePath());
+                }
+                removeDir(dir);
+            } else if (isDanglingSymlink(dir)) {
+                log("Trying to delete directory " + dir.getAbsolutePath()
+                    + " which looks like a broken symlink.",
+                    quiet ? Project.MSG_VERBOSE : verbosity);
+                if (!delete(dir)) {
+                    handle("Unable to delete directory "
+                           + dir.getAbsolutePath());
                 }
             }
         }
+        Resources resourcesToDelete = new Resources();
+        resourcesToDelete.setProject(getProject());
+        resourcesToDelete.setCache(true);
+        Resources filesetDirs = new Resources();
+        filesetDirs.setProject(getProject());
+        filesetDirs.setCache(true);
+        FileSet implicit = null;
+        if (usedMatchingTask && dir != null && dir.isDirectory()) {
+            //add the files from the default fileset:
+            implicit = getImplicitFileSet();
+            implicit.setProject(getProject());
+            filesets.add(implicit);
+        }
 
-        // delete the files from the default fileset
-        if (usedMatchingTask && dir != null) {
-            try {
-                DirectoryScanner ds = super.getDirectoryScanner(dir);
-                String[] files = ds.getIncludedFiles();
-                String[] dirs = ds.getIncludedDirectories();
-                removeFiles(dir, files, dirs);
-            } catch (BuildException be) {
-                // directory doesn't exist or is not readable
-                if (failonerror) {
-                    throw be;
-                } else {
-                    log(be.getMessage(),
-                        quiet ? Project.MSG_VERBOSE : Project.MSG_WARN);
+        for (int i = 0, size = filesets.size(); i < size; i++) {
+            FileSet fs = (FileSet) filesets.get(i);
+            if (fs.getProject() == null) {
+                log("Deleting fileset with no project specified;"
+                    + " assuming executing project", Project.MSG_VERBOSE);
+                fs = (FileSet) fs.clone();
+                fs.setProject(getProject());
+            }
+            final File fsDir = fs.getDir();
+            if (fsDir == null) {
+                throw new BuildException(
+                        "File or Resource without directory or file specified");
+            } else if (!fsDir.isDirectory()) {
+                handle("Directory does not exist: " + fsDir);
+            } else {
+                DirectoryScanner ds = fs.getDirectoryScanner();
+                // the previous line has already scanned the
+                // filesystem, in order to avoid a rescan when later
+                // iterating, capture the results now and store them
+                final String[] files = ds.getIncludedFiles();
+                resourcesToDelete.add(new ResourceCollection() {
+                        public boolean isFilesystemOnly() {
+                            return true;
+                        }
+                        public int size() {
+                            return files.length;
+                        }
+                        public Iterator iterator() {
+                            return new FileResourceIterator(getProject(),
+                                                            fsDir, files);
+                        }
+                    });
+                if (includeEmpty) {
+                    filesetDirs.add(new ReverseDirs(getProject(), fsDir,
+                                                    ds
+                                                    .getIncludedDirectories()));
                 }
+
+                if (removeNotFollowedSymlinks) {
+                    String[] n = ds.getNotFollowedSymlinks();
+                    if (n.length > 0) {
+                        String[] links = new String[n.length];
+                        System.arraycopy(n, 0, links, 0, n.length);
+                        Arrays.sort(links, ReverseDirs.REVERSE);
+                        for (int l = 0; l < links.length; l++) {
+                            try {
+                                SYMLINK_UTILS
+                                    .deleteSymbolicLink(new File(links[l]),
+                                                        this);
+                            } catch (java.io.IOException ex) {
+                                handle(ex);
+                            }
+                        }
+                    }
+                }
+            }
+        }
+        resourcesToDelete.add(filesetDirs);
+        if (rcs != null) {
+            // sort first to files, then dirs
+            Restrict exists = new Restrict();
+            exists.add(EXISTS);
+            exists.add(rcs);
+            Sort s = new Sort();
+            s.add(REVERSE_FILESYSTEM);
+            s.add(exists);
+            resourcesToDelete.add(s);
+        }
+        try {
+            if (resourcesToDelete.isFilesystemOnly()) {
+                for (Iterator iter = resourcesToDelete.iterator(); iter.hasNext();) {
+                    // nonexistent resources could only occur if we already
+                    // deleted something from a fileset:
+                    Resource r = (Resource) iter.next();
+                    File f = ((FileProvider) r.as(FileProvider.class))
+                              .getFile();
+                    if (!f.exists()) {
+                        continue;
+                    }
+                    if (!(f.isDirectory()) || f.list().length == 0) {
+                        log("Deleting " + f, verbosity);
+                        if (!delete(f) && failonerror) {
+                            handle("Unable to delete "
+                                + (f.isDirectory() ? "directory " : "file ") + f);
+                        }
+                    }
+                }
+            } else {
+                 handle(getTaskName() + " handles only filesystem resources");
+            }
+        } catch (Exception e) {
+            handle(e);
+        } finally {
+            if (implicit != null) {
+                filesets.remove(implicit);
             }
         }
     }
@@ -539,33 +696,35 @@ public class Delete extends MatchingTask {
 //************************************************************************
 //  protected and private methods
 //************************************************************************
+
+    private void handle(String msg) {
+        handle(new BuildException(msg));
+    }
+
+    private void handle(Exception e) {
+        if (failonerror) {
+            throw (e instanceof BuildException)
+                ? (BuildException) e : new BuildException(e);
+        }
+        log(e, quiet ? Project.MSG_VERBOSE : verbosity);
+    }
+
     /**
      * Accommodate Windows bug encountered in both Sun and IBM JDKs.
      * Others possible. If the delete does not work, call System.gc(),
      * wait a little and try again.
      */
     private boolean delete(File f) {
-        if (!f.delete()) {
-            if (Os.isFamily("windows")) {
-                System.gc();
-            }
-            try {
-                Thread.sleep(DELETE_RETRY_SLEEP_MILLIS);
-            } catch (InterruptedException ex) {
-                // Ignore Exception
-            }
-            if (!f.delete()) {
-                if (deleteOnExit) {
-                    int level = quiet ? Project.MSG_VERBOSE : Project.MSG_INFO;
-                    log("Failed to delete " + f + ", calling deleteOnExit."
-                        + " This attempts to delete the file when the ant jvm"
-                        + " has exited and might not succeed."
-                        , level);
-                    f.deleteOnExit();
-                    return true;
-                }
-                return false;
+        if (!FILE_UTILS.tryHardToDelete(f)) {
+            if (deleteOnExit) {
+                int level = quiet ? Project.MSG_VERBOSE : Project.MSG_INFO;
+                log("Failed to delete " + f + ", calling deleteOnExit."
+                    + " This attempts to delete the file when the Ant jvm"
+                    + " has exited and might not succeed.", level);
+                f.deleteOnExit();
+                return true;
             }
+            return false;
         }
         return true;
     }
@@ -586,29 +745,15 @@ public class Delete extends MatchingTask {
             if (f.isDirectory()) {
                 removeDir(f);
             } else {
-                log("Deleting " + f.getAbsolutePath(), verbosity);
+                log("Deleting " + f.getAbsolutePath(), quiet ? Project.MSG_VERBOSE : verbosity);
                 if (!delete(f)) {
-                    String message = "Unable to delete file "
-                        + f.getAbsolutePath();
-                    if (failonerror) {
-                        throw new BuildException(message);
-                    } else {
-                        log(message,
-                            quiet ? Project.MSG_VERBOSE : Project.MSG_WARN);
-                    }
+                    handle("Unable to delete file " + f.getAbsolutePath());
                 }
             }
         }
         log("Deleting directory " + d.getAbsolutePath(), verbosity);
         if (!delete(d)) {
-            String message = "Unable to delete directory "
-                + dir.getAbsolutePath();
-            if (failonerror) {
-                throw new BuildException(message);
-            } else {
-                log(message,
-                    quiet ? Project.MSG_VERBOSE : Project.MSG_WARN);
-            }
+            handle("Unable to delete directory " + d.getAbsolutePath());
         }
     }
 
@@ -622,19 +767,13 @@ public class Delete extends MatchingTask {
     protected void removeFiles(File d, String[] files, String[] dirs) {
         if (files.length > 0) {
             log("Deleting " + files.length + " files from "
-                + d.getAbsolutePath());
+                + d.getAbsolutePath(), quiet ? Project.MSG_VERBOSE : verbosity);
             for (int j = 0; j < files.length; j++) {
                 File f = new File(d, files[j]);
-                log("Deleting " + f.getAbsolutePath(), verbosity);
+                log("Deleting " + f.getAbsolutePath(),
+                        quiet ? Project.MSG_VERBOSE : verbosity);
                 if (!delete(f)) {
-                    String message = "Unable to delete file "
-                        + f.getAbsolutePath();
-                    if (failonerror) {
-                        throw new BuildException(message);
-                    } else {
-                        log(message,
-                            quiet ? Project.MSG_VERBOSE : Project.MSG_WARN);
-                    }
+                    handle("Unable to delete file " + f.getAbsolutePath());
                 }
             }
         }
@@ -642,19 +781,14 @@ public class Delete extends MatchingTask {
         if (dirs.length > 0 && includeEmpty) {
             int dirCount = 0;
             for (int j = dirs.length - 1; j >= 0; j--) {
-                File dir = new File(d, dirs[j]);
-                String[] dirFiles = dir.list();
+                File currDir = new File(d, dirs[j]);
+                String[] dirFiles = currDir.list();
                 if (dirFiles == null || dirFiles.length == 0) {
-                    log("Deleting " + dir.getAbsolutePath(), verbosity);
-                    if (!delete(dir)) {
-                        String message = "Unable to delete directory "
-                                + dir.getAbsolutePath();
-                        if (failonerror) {
-                            throw new BuildException(message);
-                        } else {
-                            log(message,
-                                quiet ? Project.MSG_VERBOSE : Project.MSG_WARN);
-                        }
+                    log("Deleting " + currDir.getAbsolutePath(),
+                            quiet ? Project.MSG_VERBOSE : verbosity);
+                    if (!delete(currDir)) {
+                        handle("Unable to delete directory "
+                                + currDir.getAbsolutePath());
                     } else {
                         dirCount++;
                     }
@@ -662,10 +796,23 @@ public class Delete extends MatchingTask {
             }
 
             if (dirCount > 0) {
-                log("Deleted " + dirCount + " director"
-                    + (dirCount == 1 ? "y" : "ies")
-                    + " from " + d.getAbsolutePath());
+                log("Deleted "
+                     + dirCount
+                     + " director" + (dirCount == 1 ? "y" : "ies")
+                     + " form " + d.getAbsolutePath(),
+                     quiet ? Project.MSG_VERBOSE : verbosity);
             }
         }
     }
+
+    private boolean isDanglingSymlink(File f) {
+        try {
+            return SYMLINK_UTILS.isDanglingSymbolicLink(f);
+        } catch (java.io.IOException e) {
+            log("Error while trying to detect " + f.getAbsolutePath()
+                + " as broken symbolic link. " + e.getMessage(),
+                quiet ? Project.MSG_VERBOSE : verbosity);
+            return false;
+        }
+    }
 }
