@@ -1,9 +1,10 @@
 /*
- * Copyright  2000-2004 The Apache Software Foundation
- *
- *  Licensed under the Apache License, Version 2.0 (the "License");
- *  you may not use this file except in compliance with the License.
- *  You may obtain a copy of the License at
+ *  Licensed to the Apache Software Foundation (ASF) under one or more
+ *  contributor license agreements.  See the NOTICE file distributed with
+ *  this work for additional information regarding copyright ownership.
+ *  The ASF licenses this file to You under the Apache License, Version 2.0
+ *  (the "License"); you may not use this file except in compliance with
+ *  the License.  You may obtain a copy of the License at
  *
  *      http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -19,8 +20,6 @@ package org.apache.tools.ant.taskdefs.optional;
 import java.io.File;
 import java.io.FileInputStream;
 import java.io.IOException;
-import java.net.MalformedURLException;
-import java.net.URL;
 import java.util.Vector;
 
 import org.apache.tools.ant.AntClassLoader;
@@ -35,6 +34,7 @@ import org.apache.tools.ant.types.Reference;
 import org.apache.tools.ant.types.XMLCatalog;
 import org.apache.tools.ant.util.FileUtils;
 import org.apache.tools.ant.util.JAXPUtils;
+import org.apache.tools.ant.util.XmlConstants;
 
 import org.xml.sax.EntityResolver;
 import org.xml.sax.ErrorHandler;
@@ -59,13 +59,14 @@ public class XMLValidateTask extends Task {
     /**
      * helper for path -> URI and URI -> path conversions.
      */
-    private static FileUtils fu = FileUtils.newFileUtils();
+    private static final FileUtils FILE_UTILS = FileUtils.getFileUtils();
 
     protected static final String INIT_FAILED_MSG =
         "Could not start xml validation: ";
 
     // ant task properties
     // defaults
+    // CheckStyle:VisibilityModifier OFF - bc
     protected boolean failOnError = true;
     protected boolean warn = true;
     protected boolean lenient = false;
@@ -87,6 +88,7 @@ public class XMLValidateTask extends Task {
     // XMLReader used to validation process
     protected ValidatorErrorHandler errorHandler = new ValidatorErrorHandler();
     // to report sax parsing errors
+    // CheckStyle:VisibilityModifier ON
 
     /** The vector to store all attributes (features) to be set on the parser. **/
     private Vector attributeList = new Vector();
@@ -97,6 +99,11 @@ public class XMLValidateTask extends Task {
     private final Vector propertyList = new Vector();
 
     private XMLCatalog xmlCatalog = new XMLCatalog();
+    /** Message for sucessfull validation */
+    public static final String MESSAGE_FILES_VALIDATED
+        = " file(s) have been successfully validated.";
+
+    private AntClassLoader readerLoader = null;
 
     /**
      * Specify how parser error are to be handled.
@@ -210,7 +217,9 @@ public class XMLValidateTask extends Task {
      * Add an attribute nested element. This is used for setting arbitrary
      * features of the SAX parser.
      * Valid attributes
-     * <a href="http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description">include</a>
+     * <a href=
+     * "http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description"
+     * >include</a>
      * @return attribute created
      * @since ant1.6
      */
@@ -260,19 +269,32 @@ public class XMLValidateTask extends Task {
     protected EntityResolver getEntityResolver() {
         return xmlCatalog;
     }
+
+    /**
+     * get the XML reader. Non-null only after {@link #initValidator()}.
+     * If the reader is an instance of  {@link ParserAdapter} then
+     * the parser is a SAX1 parser, and you cannot call
+     * {@link #setFeature(String, boolean)} or {@link #setProperty(String, String)}
+     * on it.
+     * @return the XML reader or null.
+     */
+    protected XMLReader getXmlReader() {
+        return xmlReader;
+    }
+
     /**
      * execute the task
      * @throws BuildException if <code>failonerror</code> is true and an error happens
      */
     public void execute() throws BuildException {
-
+        try {
         int fileProcessed = 0;
         if (file == null && (filesets.size() == 0)) {
             throw new BuildException(
                 "Specify at least one source - " + "a file or a fileset.");
         }
 
-        initValidator();
+
 
         if (file != null) {
             if (file.exists() && file.canRead() && file.isFile()) {
@@ -300,31 +322,81 @@ public class XMLValidateTask extends Task {
                 fileProcessed++;
             }
         }
-        log(fileProcessed + " file(s) have been successfully validated.");
+        onSuccessfulValidation(fileProcessed);
+        } finally {
+            cleanup();
+        }
+    }
+
+    /**
+     * handler called on successful file validation.
+     * @param fileProcessed number of files processed.
+     */
+    protected void onSuccessfulValidation(int fileProcessed) {
+        log(fileProcessed + MESSAGE_FILES_VALIDATED);
     }
 
     /**
      * init the parser :
      * load the parser class, and set features if necessary
+     * It is only after this that the reader is valid
+     * @throws BuildException if something went wrong
      */
-    private void initValidator() {
+    protected void initValidator() {
 
+        xmlReader = createXmlReader();
+
+        xmlReader.setEntityResolver(getEntityResolver());
+        xmlReader.setErrorHandler(errorHandler);
+
+        if (!isSax1Parser()) {
+            // turn validation on
+            if (!lenient) {
+                setFeature(XmlConstants.FEATURE_VALIDATION, true);
+            }
+            // set the feature from the attribute list
+            for (int i = 0; i < attributeList.size(); i++) {
+                Attribute feature = (Attribute) attributeList.elementAt(i);
+                setFeature(feature.getName(), feature.getValue());
+
+            }
+            // Sets properties
+            for (int i = 0; i < propertyList.size(); i++) {
+                final Property prop = (Property) propertyList.elementAt(i);
+                setProperty(prop.getName(), prop.getValue());
+            }
+        }
+    }
+
+    /**
+     * test that returns true if we are using a SAX1 parser.
+     * @return true when a SAX1 parser is in use
+     */
+    protected boolean isSax1Parser() {
+        return (xmlReader instanceof ParserAdapter);
+    }
+
+    /**
+     * create the XML reader.
+     * This is one by instantiating anything specified by {@link #readerClassName},
+     * falling back to a default reader if not.
+     * If the returned reader is an instance of {@link ParserAdapter} then
+     * we have created and wrapped a SAX1 parser.
+     * @return the new XMLReader.
+     */
+    protected XMLReader createXmlReader() {
         Object reader = null;
         if (readerClassName == null) {
-            try {
-                reader = JAXPUtils.getXMLReader();
-            } catch (BuildException exc) {
-                reader = JAXPUtils.getParser();
-            }
+            reader = createDefaultReaderOrParser();
         } else {
 
             Class readerClass = null;
             try {
                 // load the parser class
                 if (classpath != null) {
-                    AntClassLoader loader =
-                        getProject().createClassLoader(classpath);
-                    readerClass = Class.forName(readerClassName, true, loader);
+                    readerLoader = getProject().createClassLoader(classpath);
+                    readerClass = Class.forName(readerClassName, true,
+                                                readerLoader);
                 } else {
                     readerClass = Class.forName(readerClassName);
                 }
@@ -340,8 +412,9 @@ public class XMLValidateTask extends Task {
         }
 
         // then check it implements XMLReader
+        XMLReader newReader;
         if (reader instanceof XMLReader) {
-            xmlReader = (XMLReader) reader;
+            newReader = (XMLReader) reader;
             log(
                 "Using SAX2 reader " + reader.getClass().getName(),
                 Project.MSG_VERBOSE);
@@ -349,7 +422,7 @@ public class XMLValidateTask extends Task {
 
             // see if it is a SAX1 Parser
             if (reader instanceof Parser) {
-                xmlReader = new ParserAdapter((Parser) reader);
+                newReader = new ParserAdapter((Parser) reader);
                 log(
                     "Using SAX1 parser " + reader.getClass().getName(),
                     Project.MSG_VERBOSE);
@@ -360,38 +433,55 @@ public class XMLValidateTask extends Task {
                         + " implements nor SAX1 Parser nor SAX2 XMLReader.");
             }
         }
+        return newReader;
+    }
 
-        xmlReader.setEntityResolver(getEntityResolver());
-        xmlReader.setErrorHandler(errorHandler);
-
-        if (!(xmlReader instanceof ParserAdapter)) {
-            // turn validation on
-            if (!lenient) {
-                setFeature("http://xml.org/sax/features/validation", true);
-            }
-            // set the feature from the attribute list
-            for (int i = 0; i < attributeList.size(); i++) {
-                Attribute feature = (Attribute) attributeList.elementAt(i);
-                setFeature(feature.getName(), feature.getValue());
-
-            }
-
-            // Sets properties
-            for (int i = 0; i < propertyList.size(); i++) {
-                final Property prop = (Property) propertyList.elementAt(i);
-                setProperty(prop.getName(), prop.getValue());
-            }
+    /**
+     * Cleans up resources.
+     *
+     * @since Ant 1.8.0
+     */
+    protected void cleanup() {
+        if (readerLoader != null) {
+            readerLoader.cleanup();
+            readerLoader = null;
         }
     }
 
+    /**
+     *
+     * @return
+     */
+    private Object createDefaultReaderOrParser() {
+        Object reader;
+        try {
+            reader = createDefaultReader();
+        } catch (BuildException exc) {
+            reader = JAXPUtils.getParser();
+        }
+        return reader;
+    }
+
+    /**
+     * create a reader if the use of the class did not specify another one.
+     * If a BuildException is thrown, the caller may revert to an alternate
+     * reader.
+     * @return a new reader.
+     * @throws BuildException if something went wrong
+     */
+    protected XMLReader createDefaultReader() {
+        return JAXPUtils.getXMLReader();
+    }
+
     /**
      * Set a feature on the parser.
      * @param feature the name of the feature to set
      * @param value the value of the feature
+     * @throws BuildException if the feature was not supported
      */
-    private void setFeature(String feature, boolean value)
+    protected void setFeature(String feature, boolean value)
         throws BuildException {
-
+        log("Setting feature " + feature + "=" + value, Project.MSG_DEBUG);
         try {
             xmlReader.setFeature(feature, value);
         } catch (SAXNotRecognizedException e) {
@@ -419,8 +509,9 @@ public class XMLValidateTask extends Task {
      * @param name a property name
      * @param value a property value.
      * @throws BuildException if an error occurs.
+     * @throws BuildException if the property was not supported
      */
-    private void setProperty(String name, String value) throws BuildException {
+    protected void setProperty(String name, String value) throws BuildException {
         // Validates property
         if (name == null || value == null) {
             throw new BuildException("Property name and value must be specified.");
@@ -449,36 +540,42 @@ public class XMLValidateTask extends Task {
 
     /**
      * parse the file
+     * @param afile the file to validate.
+     * @return true if the file validates.
      */
-    private void doValidate(File afile) {
+    protected boolean doValidate(File afile) {
+        //for every file, we have a new instance of the validator
+        initValidator();
+        boolean result = true;
         try {
             log("Validating " + afile.getName() + "... ", Project.MSG_VERBOSE);
             errorHandler.init(afile);
             InputSource is = new InputSource(new FileInputStream(afile));
-            String uri = fu.toURI(afile.getAbsolutePath());
+            String uri = FILE_UTILS.toURI(afile.getAbsolutePath());
             is.setSystemId(uri);
             xmlReader.parse(is);
         } catch (SAXException ex) {
+            log("Caught when validating: " + ex.toString(), Project.MSG_DEBUG);
             if (failOnError) {
                 throw new BuildException(
                     "Could not validate document " + afile);
-            } else {
-                log("Could not validate document " + afile + ": " + ex.toString());
             }
+            log("Could not validate document " + afile + ": " + ex.toString());
+            result = false;
         } catch (IOException ex) {
             throw new BuildException(
                 "Could not validate document " + afile,
                 ex);
         }
-
         if (errorHandler.getFailure()) {
             if (failOnError) {
                 throw new BuildException(
                     afile + " is not a valid XML document.");
-            } else {
-                log(afile + " is not a valid XML document", Project.MSG_ERR);
             }
+            result = false;
+            log(afile + " is not a valid XML document", Project.MSG_ERR);
         }
+        return result;
     }
 
     /**
@@ -490,9 +587,11 @@ public class XMLValidateTask extends Task {
      */
     protected class ValidatorErrorHandler implements ErrorHandler {
 
+        // CheckStyle:VisibilityModifier OFF - bc
         protected File currentFile = null;
         protected String lastErrorMessage = null;
         protected boolean failed = false;
+        // CheckStyle:VisibilityModifier ON
         /**
          * initialises the class
          * @param file file used
@@ -545,18 +644,22 @@ public class XMLValidateTask extends Task {
         private String getMessage(SAXParseException e) {
             String sysID = e.getSystemId();
             if (sysID != null) {
-                try {
-                    int line = e.getLineNumber();
-                    int col = e.getColumnNumber();
-                    return new URL(sysID).getFile()
-                        + (line == -1
-                            ? ""
-                            : (":" + line + (col == -1 ? "" : (":" + col))))
-                        + ": "
-                        + e.getMessage();
-                } catch (MalformedURLException mfue) {
-                    // ignore and just return exception message
+                String name = sysID;
+                if (sysID.startsWith("file:")) {
+                    try {
+                        name = FILE_UTILS.fromURI(sysID);
+                    } catch (Exception ex) {
+                        // if this is not a valid file: just use the uri
+                    }
                 }
+                int line = e.getLineNumber();
+                int col = e.getColumnNumber();
+                return  name
+                    + (line == -1
+                       ? ""
+                       : (":" + line + (col == -1 ? "" : (":" + col))))
+                    + ": "
+                    + e.getMessage();
             }
             return e.getMessage();
         }
@@ -566,10 +669,12 @@ public class XMLValidateTask extends Task {
      * The class to create to set a feature of the parser.
      * @since ant1.6
      */
-    public class Attribute {
+    public static class Attribute {
         /** The name of the attribute to set.
          *
-         * Valid attributes <a href="http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description">include.</a>
+         * Valid attributes <a href=
+         * "http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description"
+         * >include.</a>
          */
         private String attributeName = null;
 
@@ -616,7 +721,7 @@ public class XMLValidateTask extends Task {
      * XML parser properties</a> for usable properties
      * @since ant 1.6.2
      */
-    public final class Property {
+    public static final class Property {
 
         private String name;
         private String value;
@@ -652,4 +757,6 @@ public class XMLValidateTask extends Task {
 
     } // Property
 
+
+
 }
